import typing

from graphql import GraphQLUnionType

from strawberry.exceptions import UnallowedReturnTypeForUnion, WrongReturnTypeForUnion
from strawberry.type import TypeDefinition
from strawberry.union import StrawberryUnion
from strawberry.utils.typing import (
    get_list_annotation,
    is_generic,
    is_list,
    is_type_var,
)

from .types import ConcreteType, TypeMap


def _get_type_mapping_from_actual_type(root) -> typing.Dict[typing.Any, typing.Type]:
    # we map ~T to the actual type of root
    type_var_to_actual_type = {}

    for field_name, annotation in root.__annotations__.items():
        # when we have a list we want to get the type of the elements contained in the
        # list, to do so we currently only get the first time (if the list is not empty)
        # this might break in more complex cases, but should suffice for now.

        if is_list(annotation):
            annotation = get_list_annotation(annotation)

            if is_type_var(annotation):
                values = getattr(root, field_name)

                if values:
                    type_var_to_actual_type[annotation] = type(values[0])

        elif is_type_var(annotation):
            type_var_to_actual_type[annotation] = type(getattr(root, field_name))

        elif is_generic(annotation):
            type_var_to_actual_type.update(
                _get_type_mapping_from_actual_type(getattr(root, field_name))
            )

    return type_var_to_actual_type


def _find_type_for_generic_union(root: typing.Any) -> TypeDefinition:
    # this is a ordered tuple of the type vars for the generic class, so for
    # typing.Generic[T, V] it would return (T, V)
    type_params = root.__parameters__

    mapping = _get_type_mapping_from_actual_type(root)

    if not mapping:
        # if we weren't able to find a mapping, ie. when returning an empty list
        # for a generic type, then we fall back to returning the first copy.
        # This a very simplistic heuristic and it is bound to break with complex
        # uses cases. We can improve it later if this becomes an issue.

        return next((t for t in root._copies.values()))._type_definition

    types = tuple(mapping[param] for param in type_params)

    type = root._copies.get(types)

    if type is None:
        raise ValueError(f"Unable to find type for {root.__class__} and {types}")

    return type._type_definition


def get_union_type(
    union_definition: StrawberryUnion, type_map: TypeMap
) -> GraphQLUnionType:
    from .object_type import get_object_type

    def _resolve_type(root, info, _type):
        if not hasattr(root, "_type_definition"):
            raise WrongReturnTypeForUnion(info.field_name, str(type(root)))

        type_definition = root._type_definition

        if is_generic(type(root)):
            type_definition = _find_type_for_generic_union(root)

        returned_type = type_map[type_definition.name].implementation

        if returned_type not in _type.types:
            raise UnallowedReturnTypeForUnion(
                info.field_name, str(type(root)), _type.types
            )

        return returned_type

    types = union_definition.types

    if union_definition.name not in type_map:
        type_map[union_definition.name] = ConcreteType(
            definition=union_definition,
            implementation=GraphQLUnionType(
                union_definition.name,
                [get_object_type(type, type_map) for type in types],
                description=union_definition.description,
                resolve_type=_resolve_type,
            ),
        )

    return typing.cast(GraphQLUnionType, type_map[union_definition.name].implementation)



import dataclasses
from typing import Callable, List, Optional, Type, Union

from .permission import BasePermission
from .types.fields.resolver import StrawberryResolver
from .types.types import FederationFieldParams, FieldDefinition
from .utils.str_converters import to_camel_case


_RESOLVER_TYPE = Union[StrawberryResolver, Callable]


class StrawberryField(dataclasses.Field):
    _field_definition: FieldDefinition

    def __init__(self, field_definition: FieldDefinition):
        self._field_definition = field_definition

        super().__init__(  # type: ignore
            default=dataclasses.MISSING,
            default_factory=dataclasses.MISSING,
            init=field_definition.base_resolver is None,
            repr=True,
            hash=None,
            compare=True,
            metadata=None,
        )

    def __call__(self, resolver: _RESOLVER_TYPE) -> "StrawberryField":
        """Add a resolver to the field"""

        # Allow for StrawberryResolvers or bare functions to be provided
        if not isinstance(resolver, StrawberryResolver):
            resolver = StrawberryResolver(resolver)

        self._field_definition.origin_name = resolver.name
        self._field_definition.origin = resolver.wrapped_func
        self._field_definition.base_resolver = resolver
        self._field_definition.arguments = resolver.arguments
        self._field_definition.type = resolver.type

        # Don't add field to __init__ or __repr__
        self.init = False
        self.repr = False

        return self

    def __setattr__(self, name, value):
        if name == "type":
            self._field_definition.type = value

        if value and name == "name":
            if not self._field_definition.origin_name:
                self._field_definition.origin_name = value

            if not self._field_definition.name:
                self._field_definition.name = to_camel_case(value)

        return super().__setattr__(name, value)


def field(
    resolver: Optional[_RESOLVER_TYPE] = None,
    *,
    name: Optional[str] = None,
    is_subscription: bool = False,
    description: Optional[str] = None,
    permission_classes: Optional[List[Type[BasePermission]]] = None,
    federation: Optional[FederationFieldParams] = None,
    deprecation_reason: Optional[str] = None,
) -> StrawberryField:
    """Annotates a method or property as a GraphQL field.

    This is normally used inside a type declaration:

    >>> @strawberry.type:
    >>> class X:
    >>>     field_abc: str = strawberry.field(description="ABC")

    >>>     @strawberry.field(description="ABC")
    >>>     def field_with_resolver(self, info) -> str:
    >>>         return "abc"

    it can be used both as decorator and as a normal function.
    """

    field_definition = FieldDefinition(
        origin_name=None,  # modified by resolver in __call__
        name=name,
        type=None,
        description=description,
        is_subscription=is_subscription,
        permission_classes=permission_classes or [],
        arguments=[],  # modified by resolver in __call__
        federation=federation or FederationFieldParams(),
        deprecation_reason=deprecation_reason,
    )

    field_ = StrawberryField(field_definition)

    if resolver:
        return field_(resolver)
    return field_


__all__ = ["FederationFieldParams", "FieldDefinition", "StrawberryField", "field"]



from dataclasses import dataclass
from typing import Callable, Dict, Optional, Type

from .exceptions import ScalarAlreadyRegisteredError
from .utils.str_converters import to_camel_case


def identity(x):
    return x


@dataclass
class ScalarDefinition:
    name: str
    description: Optional[str]
    serialize: Optional[Callable]
    parse_value: Optional[Callable]
    parse_literal: Optional[Callable]


SCALAR_REGISTRY: Dict[Type, ScalarDefinition] = {}


class ScalarWrapper:
    _scalar_definition: ScalarDefinition

    def __init__(self, wrap):
        self.wrap = wrap

    def __call__(self, *args, **kwargs):
        return self.wrap(*args, **kwargs)


def _process_scalar(
    cls,
    *,
    name: str = None,
    description: str = None,
    serialize: Callable = None,
    parse_value: Callable = None,
    parse_literal: Callable = None
):

    name = name or to_camel_case(cls.__name__)

    if cls in SCALAR_REGISTRY:
        raise ScalarAlreadyRegisteredError(name)

    wrapper = ScalarWrapper(cls)
    wrapper._scalar_definition = ScalarDefinition(
        name=name,
        description=description,
        serialize=serialize,
        parse_literal=parse_literal,
        parse_value=parse_value,
    )

    SCALAR_REGISTRY[cls] = wrapper._scalar_definition

    return wrapper


def scalar(
    cls=None,
    *,
    name: str = None,
    description: str = None,
    serialize: Callable = identity,
    parse_value: Optional[Callable] = None,
    parse_literal: Optional[Callable] = None
):
    """Annotates a class or type as a GraphQL custom scalar.

    Example usages:

    >>> strawberry.scalar(
    >>>     datetime.date,
    >>>     serialize=lambda value: value.isoformat(),
    >>>     parse_value=datetime.parse_date
    >>> )

    >>> Base64Encoded = strawberry.scalar(
    >>>     NewType("Base64Encoded", bytes),
    >>>     serialize=base64.b64encode,
    >>>     parse_value=base64.b64decode
    >>> )

    >>> @strawberry.scalar(
    >>>     serialize=lambda value: ",".join(value.items),
    >>>     parse_value=lambda value: CustomList(value.split(","))
    >>> )
    >>> class CustomList:
    >>>     def __init__(self, items):
    >>>         self.items = items

    """

    if parse_value is None:
        parse_value = cls

    def wrap(cls):
        return _process_scalar(
            cls,
            name=name,
            description=description,
            serialize=serialize,
            parse_value=parse_value,
            parse_literal=parse_literal,
        )

    if cls is None:
        return wrap

    return wrap(cls)



import pathlib


def get_graphiql_html() -> str:
    here = pathlib.Path(__file__).parents[1]
    path = here / "static/graphiql.html"

    with open(path) as f:
        template = f.read()

    return template.replace("{{ SUBSCRIPTION_ENABLED }}", "true")



from typing import Dict, List, cast

from graphql import GraphQLArgument, Undefined
from graphql.type.definition import GraphQLInputType

from strawberry.arguments import ArgumentDefinition, undefined

from .type import get_graphql_type
from .types import TypeMap


def convert_argument(
    argument: ArgumentDefinition, type_map: TypeMap
) -> GraphQLArgument:
    # TODO: test and support generic arguments?
    default_value = (
        Undefined if argument.default_value is undefined else argument.default_value
    )

    # TODO: we could overload the function to tell mypy that it returns input types too
    argument_type = cast(GraphQLInputType, get_graphql_type(argument, type_map))

    return GraphQLArgument(
        argument_type,
        default_value=default_value,
        description=argument.description,
    )


def convert_arguments(
    arguments: List[ArgumentDefinition], type_map: TypeMap
) -> Dict[str, GraphQLArgument]:
    arguments_dict = {}

    for argument in arguments:
        name = cast(str, argument.name)
        arguments_dict[name] = convert_argument(argument, type_map)

    return arguments_dict



from .directives import get_directive_type as get_directive_type
from .object_type import get_object_type as get_object_type
from .types import ConcreteType as ConcreteType


__all__ = ["ConcreteType", "get_directive_type", "get_object_type"]



from typing import Type

from pydantic import BaseModel


class MissingFieldsListError(Exception):
    def __init__(self, type: Type[BaseModel]):
        message = f"List of fields to copy from {type} is empty"

        super().__init__(message)


class UnsupportedTypeError(Exception):
    pass


class UnregisteredTypeException(Exception):
    def __init__(self, type: BaseModel):
        message = (
            f"Cannot find a Strawberry Type for {type} did you forget to register it?"
        )

        super().__init__(message)



from abc import abstractmethod
from typing import Any, Dict, List

from typing_extensions import Protocol

from .directive import DirectiveDefinition


SPECIFIED_DIRECTIVES = {"include", "skip"}


class Middleware(Protocol):
    @abstractmethod
    def resolve(self, next_, root, info, **kwargs):
        raise NotImplementedError


class DirectivesMiddleware:
    def __init__(self, directives: List[Any]):
        self.directives: Dict[str, DirectiveDefinition] = {
            directive.directive_definition.name: directive.directive_definition
            for directive in directives
        }

    def resolve(self, next_, root, info, **kwargs):
        result = next_(root, info, **kwargs)

        for directive in info.field_nodes[0].directives:
            directive_name = directive.name.value

            if directive_name in SPECIFIED_DIRECTIVES:
                continue

            func = self.directives[directive_name].resolver

            # TODO: support converting lists

            arguments = {
                argument.name.value: argument.value.value
                for argument in directive.arguments
            }

            result = func(result, **arguments)

        return result



from typing import Any, Dict, List, Optional, Sequence, Type, Union

from graphql import (
    ExecutionContext as GraphQLExecutionContext,
    GraphQLSchema,
    get_introspection_query,
    parse,
    validate_schema,
)
from graphql.subscription import subscribe
from graphql.type.directives import specified_directives

from strawberry.custom_scalar import ScalarDefinition
from strawberry.enum import EnumDefinition
from strawberry.extensions import Extension
from strawberry.types.types import TypeDefinition
from strawberry.union import StrawberryUnion

from ..middleware import DirectivesMiddleware, Middleware
from ..printer import print_schema
from .base import ExecutionResult
from .execute import execute, execute_sync
from .types import ConcreteType, get_directive_type, get_object_type


class Schema:
    def __init__(
        self,
        # TODO: can we make sure we only allow to pass something that has been decorated?
        query: Type,
        mutation: Optional[Type] = None,
        subscription: Optional[Type] = None,
        directives=(),
        types=(),
        extensions: Sequence[Type[Extension]] = (),
        execution_context_class: Optional[Type[GraphQLExecutionContext]] = None,
    ):
        self.extensions = extensions
        self.type_map: Dict[str, ConcreteType] = {}
        self.execution_context_class = execution_context_class

        query_type = get_object_type(query, self.type_map)
        mutation_type = get_object_type(mutation, self.type_map) if mutation else None
        subscription_type = (
            get_object_type(subscription, self.type_map) if subscription else None
        )

        self.middleware: List[Middleware] = [DirectivesMiddleware(directives)]

        directives = [
            get_directive_type(directive, self.type_map) for directive in directives
        ]

        self._schema = GraphQLSchema(
            query=query_type,
            mutation=mutation_type,
            subscription=subscription_type if subscription else None,
            directives=specified_directives + directives,
            types=[get_object_type(type, self.type_map) for type in types],
        )

        # Validate schema early because we want developers to know about
        # possible issues as soon as possible
        errors = validate_schema(self._schema)
        if errors:
            formatted_errors = "\n\n".join(f"❌ {error.message}" for error in errors)
            raise ValueError(f"Invalid Schema. Errors:\n\n{formatted_errors}")

        self.query = self.type_map[query_type.name]

    def get_type_by_name(
        self, name: str
    ) -> Optional[
        Union[TypeDefinition, ScalarDefinition, EnumDefinition, StrawberryUnion]
    ]:
        if name in self.type_map:
            return self.type_map[name].definition

        return None

    async def execute(
        self,
        query: str,
        variable_values: Optional[Dict[str, Any]] = None,
        context_value: Optional[Any] = None,
        root_value: Optional[Any] = None,
        operation_name: Optional[str] = None,
    ) -> ExecutionResult:
        result = await execute(
            self._schema,
            query,
            variable_values=variable_values,
            root_value=root_value,
            context_value=context_value,
            operation_name=operation_name,
            additional_middlewares=self.middleware,
            extensions=self.extensions,
            execution_context_class=self.execution_context_class,
        )

        return ExecutionResult(
            data=result.data,
            errors=result.errors,
            extensions=result.extensions,
        )

    def execute_sync(
        self,
        query: str,
        variable_values: Optional[Dict[str, Any]] = None,
        context_value: Optional[Any] = None,
        root_value: Optional[Any] = None,
        operation_name: Optional[str] = None,
    ) -> ExecutionResult:
        result = execute_sync(
            self._schema,
            query,
            variable_values=variable_values,
            root_value=root_value,
            context_value=context_value,
            operation_name=operation_name,
            additional_middlewares=self.middleware,
            extensions=self.extensions,
            execution_context_class=self.execution_context_class,
        )

        return ExecutionResult(
            data=result.data,
            errors=result.errors,
            extensions=result.extensions,
        )

    async def subscribe(
        self,
        query: str,
        variable_values: Optional[Dict[str, Any]] = None,
        context_value: Optional[Any] = None,
        root_value: Optional[Any] = None,
        operation_name: Optional[str] = None,
    ):
        return await subscribe(
            self._schema,
            parse(query),
            root_value=root_value,
            context_value=context_value,
            variable_values=variable_values,
            operation_name=operation_name,
        )

    def as_str(self) -> str:
        return print_schema(self)

    __str__ = as_str

    def introspect(self) -> Dict[str, Any]:
        """Return the introspection query result for the current schema

        Raises:
            ValueError: If the introspection query fails due to an invalid schema
        """
        introspection = self.execute_sync(get_introspection_query())
        if introspection.errors or not introspection.data:
            raise ValueError(f"Invalid Schema. Errors {introspection.errors!r}")

        return introspection.data
